import json
import base64
import urllib3

from .utils import expand_filters, calculate_age
from prettytable import PrettyTable


def get_allocatable_gpus(nod):
    if 'nvidia.com/gpu' in nod.status.capacity:
        return nod.status.capacity['nvidia.com/gpu']

    return ''


def format_bytes(size):
    # 2**10 = 1024
    if 'Ki' in size:
        size = int(str(size).replace('Ki', '')) * 1024

    power = 2**10
    n = 0
    power_labels = {0 : '', 1: 'K', 2: 'M', 3: 'G', 4: 'T'}
    while size > power:
        size /= power
        n += 1
    return '{0} {1}'.format(round(size, 2), power_labels[n]+'B')


def get_instance_type(metadata):
    if 'node.kubernetes.io/instance-type' in metadata.labels.keys():
        return metadata.labels['node.kubernetes.io/instance-type']

    return ''


def get_node_group_name(metadata):
    if 'node.kubernetes.io/nodegroup' in metadata.labels.keys():
        return metadata.labels['node.kubernetes.io/nodegroup']

    return ''


def get_node_taints(node_spec):
    taints = []
    if node_spec.taints is None:
        return ''

    for taint in node_spec.taints:
        taints.append('{0}={1}:{2}'.format(taint.key, taint.value, taint.effect))

    return ','.join(taints)


def list_nodes_info(k8sapi, filter_criteria):
    try:
        ret = k8sapi.list_node()
    except urllib3.exceptions.MaxRetryError:
        return None

    table = PrettyTable()
    table.field_names = ['Name', 'Type', 'Group Name', 'GPUs', 'CPUs', 'Memory', 'SSD', 'Age', 'Taints']
    table.align['Name'] = 'l'
    table.align['Group Name'] = 'l'
    table.align['Memory'] = 'r'
    table.align['SSD'] = 'r'
    table.align['GPUs'] = 'r'
    table.align['CPUs'] = 'r'
    table.align['Age'] = 'r'
    table.align['Taints'] = 'l'

    query_filters = expand_filters(filter_criteria)

    for nod in ret.items:
        # print(nod)
        # exit(0)
        instance_group_name = get_node_group_name(nod.metadata)
        instance_type = get_instance_type(nod.metadata)
        instance_taints = get_node_taints(nod.spec)

        if query_filters:
            if 'name' in query_filters and query_filters['name'] not in nod.metadata.name.lower():
                continue
            if 'group' in query_filters and query_filters['group'] not in instance_group_name.lower():
                continue
            if 'type' in query_filters and query_filters['type'] not in instance_type.lower():
                continue
            if 'taints' in query_filters and query_filters['taints'] not in instance_taints.lower():
                continue

        table.add_row([
            nod.metadata.name,
            instance_type,
            instance_group_name,
            get_allocatable_gpus(nod),
            nod.status.capacity['cpu'],
            format_bytes(nod.status.capacity['memory']),
            format_bytes(nod.status.capacity['ephemeral-storage']),
            calculate_age(nod.metadata.creation_timestamp),
            instance_taints
        ])

    print(table.get_string())


class SecretsQuery(object):
    def list_secrets(self, subprocess, extra_options):
        output = subprocess.Popen([f"kubectl get secrets -o json {extra_options}"], shell=True,
                                  stdout=subprocess.PIPE).stdout.read().decode('utf-8')

        if output == '' or output is None:
            print('Could not retrieve secrets')
            exit(1)

        data = json.loads(output)

        table = PrettyTable()
        table.field_names = ['Name', 'Namespace', 'Type', 'Data']
        table.align['Name'] = 'l'
        table.align['Namespace'] = 'l'
        table.align['Type'] = 'l'
        table.align['Data'] = 'l'

        for item in data['items']:
            item_secrets = self.get_secrets_data(item)
            table.add_row([
                item['metadata']['name'],
                item['metadata']['namespace'],
                item['type'],
                str(', '.join(item_secrets))[:150],
            ])

        print(table.get_string())

    def show_secret(self, subprocess, secret_name, extra_options):
        output = subprocess.Popen([f"kubectl get -o json {secret_name} {extra_options}"], shell=True,
                                  stdout=subprocess.PIPE).stdout.read().decode('utf-8')

        try:
            data = json.loads(output)
        except json.JSONDecodeError:
            print("No secret found")
            return None

        secret_data = self.get_secrets_data(data, True)

        table = PrettyTable()
        table.field_names = ['Key', 'Value']
        table.align['Key'] = 'l'
        table.align['Value'] = 'l'

        table.add_rows([
            ['name', data['metadata']['name']],
            ['namespace', data['metadata']['namespace']],
            ['type', data['type']],
            ['---------', '---------']
        ])

        for item in secret_data:
            table.add_row([item, secret_data[item]])

        print(table.get_string())

    @staticmethod
    def get_secrets_data(item, show_values=False):
        output = {}
        if 'stringData' in item:
            for key, value in item['stringData'].items():
                if show_values:
                    output.update({key: value})
                else:
                    output.update({key: '***'})

        if 'data' in item:
            for key, value in item['data'].items():
                if show_values:
                    output.update({key: base64.b64decode(value).decode('utf-8').replace('\n', '')})
                else:
                    output.update({key: '***'})

        return output



#!/usr/bin/env python3

import typer
import boto3
import subprocess

from typing import Optional
from typing_extensions import Annotated
from importlib.metadata import version

from .aws_config import AwsConfigManager
from .aws_ec2 import AwsResourceQuery
from .aws_ecr import ContainerRegistry
from .aws_vpc import VpcManager
from .aws_route53 import Route53
from .aws_lb import AwsLb
from .k8s import list_nodes_info, SecretsQuery

app = typer.Typer(help="KPX application for aws profiles", no_args_is_help=True)


@app.command()
def cfg(
    aws_profile: Optional[str] = typer.Argument('default'),
    region: Optional[str] = typer.Argument("us-east-1"),
    output: Optional[str] = typer.Argument("json")
):
    """
    Configure ~/.aws/config file with profiles settings
    """
    AwsConfigManager.generate_config(aws_profile, region, output)


@app.command()
def cred(
    aws_profile: Optional[str] = typer.Argument('default'),
    key: Optional[str] = typer.Argument(""),
    secret: Optional[str] = typer.Argument("")
):
    """
    Configure ~/.aws/credentials file with aws credentials
    """
    AwsConfigManager.generate_credentials(aws_profile, key, secret)


@app.command()
def r53(
    zone_id: Optional[str] = typer.Argument(''),
    filter: Optional[str] = typer.Argument(''),
):
    """
    List Route53 hosted zones
    """
    client = boto3.client('route53')
    Route53.list_r53(client, zone_id, filter)


@app.command()
def ecr(repo_name: Optional[str] = typer.Argument('')):
    """
    List ECR repositories
    """
    client = boto3.client('ecr')

    ecr = ContainerRegistry(client)

    if repo_name != '':
        ecr.get_images(repo_name, AwsConfigManager.account_id)
    else:
        ecr.get_repositories()


@app.command()
def ec2(filter: Annotated[str, typer.Option('--filter', '-f')] = ''):
    """
    List EC2 instances
    """
    client = boto3.client('ec2')
    arq = AwsResourceQuery(client)
    arq.list_ec2(filter)


@app.command()
def lb():
    """
    List Load Balancers
    """
    client = boto3.client('elbv2')
    awslb = AwsLb(client)
    awslb.get_load_balancers()


@app.command()
def vpc(vpc_id: Optional[str] = typer.Argument('')):
    """
    List VPCs and show Cidr blocks or subnets for provided VPC ID
    """
    session = boto3.Session(region_name=str(AwsConfigManager.aws_region()))
    vm = VpcManager(session)
    if vpc_id != '':
        print(f'Showing item: {vpc_id}')
        vm.get_vpc_subnets(vpc_id)
    else:
        print(f'List all available VPCs in {AwsConfigManager.aws_region()}')
        client = boto3.client('ec2', region_name=AwsConfigManager.aws_region())
        vm.get_vpc_list(client)


@app.command(name="sec", hidden=True)
@app.command(name="secret")
def sec(
    secret: Optional[str] = typer.Argument(''),
    all_namespaces: Optional[bool] = typer.Option(
        None,
        '--all-namespaces',
        "-a",
        help="Show K8S secrets in all namespaces",
        is_flag=True
    ),
    select_namespace: Optional[str] = typer.Option(
        None,
        '--namespace',
        "-n",
        help="Select namespace",
    )
):
    """
    List secrets and show values
    """
    extra_options = ""
    if all_namespaces:
        extra_options += "-A"
    if select_namespace:
        extra_options += f"-n {select_namespace}"

    sq = SecretsQuery()

    if secret:
        if 'secret/' not in secret:
            secret = f'secret/{secret}'

        sq.show_secret(subprocess, secret, extra_options)

        return None

    sq.list_secrets(subprocess, extra_options)


@app.command(name='n', hidden=True)
@app.command()
def nodes(filter: Annotated[str, typer.Option('--filter', '-f')] = ''):
    """
    List EKS nodes and show some information about them
    """
    from kubernetes import client, config
    config.load_kube_config()

    k8sapi = client.CoreV1Api()

    list_nodes_info(k8sapi, filter)


@app.command(name='v', hidden=True)
@app.command(name="version")
def show_version():
    """Show current CLI tool version"""
    version_number = version("kpx")
    typer.echo(version_number)


if __name__ == '__main__':  # pragma: no cover
    app()



from prettytable import PrettyTable


class AwsLb(object):
    def __init__(self, client):
        self.client = client

    def get_load_balancers(self):
        response = self.client.describe_load_balancers(
            PageSize=400
        )

        table = PrettyTable()
        table.field_names = ['LoadBalancerName', 'VPC Id', 'State', 'Type', 'DNS Name']
        table.align['LoadBalancerName'] = 'l'
        table.align['DNS Name'] = 'r'

        for lb in response['LoadBalancers']:
            table.add_row([
                lb['LoadBalancerName'],
                lb['VpcId'],
                lb['State']['Code'],
                lb['Type'],
                lb['DNSName'],
            ])
        print(table)



import configparser
import os
import boto3

from os.path import expanduser
from .utils import Output, IniUtils


class AwsConfigManager:
    def __init__(self, file_credentials, file_config):
        self.file_credentials = file_credentials
        self.file_config = file_config

        self.creds = configparser.ConfigParser()
        self.creds.read(file_credentials)

        self.cfg = configparser.ConfigParser()
        self.cfg.read(file_config)

    @staticmethod
    def aws_region():
        return os.getenv('AWS_REGION', 'us-east-1')

    @staticmethod
    def account_id():
        client = boto3.client('sts')
        accId = client.get_caller_identity()['Account']

        return accId

    def update_credentials(self, profile, access_key, secret_key):
        if profile not in self.creds:
            self.creds.update({profile: {
                'aws_access_key_id': '',
                'aws_secret_access_key': '',
            }})

        for key in self.creds[profile]:
            new_value = ''
            if key == 'aws_access_key_id' and access_key is not None:
                new_value = access_key
            if key == 'aws_secret_access_key' and secret_key is not None:
                new_value = secret_key

            self.creds[profile][key] = new_value

        return self

    def update_config(self, profile, region, output):
        if profile != 'default':
            profile = f'profile {profile}'

        self.cfg.update({
            profile: {
                'region': region,
                'output': output,
            }
        })

    def get_credentials(self, profile_name):
        data = {}
        if profile_name in self.creds:
            for key in self.creds[profile_name]:
                data.update({key: self.creds[profile_name][key]})

        return data

    def get_config(self, profile_name):
        data = {}
        profile_string = f'profile {profile_name}' if profile_name != 'default' else 'default'
        if profile_string in self.cfg:
            for key in self.cfg[profile_string]:
                data.update({key: self.cfg[profile_string][key]})

        return data

    def write_credentials_file(self):
        with open(self.file_credentials, 'w') as file:
            self.creds.write(file)

        return self

    def write_config_file(self):
        with open(self.file_config, 'w') as file:
            self.cfg.write(file)

        return self

    @staticmethod
    def generate_config(aws_profile, region, output):
        Output.header('Updating ~/.aws/config file')
        user_home_directory = expanduser('~')

        awc = AwsConfigManager(
            f'{user_home_directory}/.aws/credentials',
            f'{user_home_directory}/.aws/config',
        )

        IniUtils.check_directory_exists(f'{user_home_directory}/.aws/')

        awc.update_config(aws_profile, region, output)
        awc.write_config_file()

    @staticmethod
    def generate_credentials(aws_profile, key, secret):
        Output.header('Updating ~/.aws/credentials file')
        user_home_directory = expanduser('~')

        awc = AwsConfigManager(
            f'{user_home_directory}/.aws/credentials',
            f'{user_home_directory}/.aws/config',
        )

        IniUtils.check_directory_exists(f'{user_home_directory}/.aws/')

        awc.update_credentials(aws_profile, key, secret)
        awc.write_credentials_file()



from .aws_ec2 import AwsResourceQuery
from prettytable import PrettyTable


class VpcManager:
    def __init__(self, session):
        self.session = session

    def get_vpc_list(self, client):
        vpcs = client.describe_vpcs()

        table = PrettyTable()
        table.field_names = ['VPC Id', 'Name', 'Cidr']
        table.align['Name'] = 'l'

        for item in vpcs['Vpcs']:
            table.add_row([
                item['VpcId'],
                AwsResourceQuery.get_instance_name_from_tags(item['Tags']) if 'Tags' in item else '-',
                item['CidrBlock']
            ])

        print(table.get_string())

    def get_vpc_subnets(self, vpc_id):
        ec2 = self.session.client('ec2')

        subnets = ec2.describe_subnets(
            Filters=[
                {
                    'Name': 'vpc-id',
                    'Values': [
                        vpc_id,
                    ],
                },
            ],
        )

        table = PrettyTable()
        table.field_names = ['Subnet ID', 'Name', 'Cidr', 'Free IPs', 'AZ']
        table.align['Subnet ID'] = 'l'
        table.align['Cidr'] = 'r'
        table.align['Freep IPs'] = 'r'

        for sn in subnets['Subnets']:
            table.add_row([
                sn['SubnetId'],
                AwsResourceQuery.get_instance_name_from_tags(sn['Tags']) if 'Tags' in sn else '-',
                sn['CidrBlock'],
                sn['AvailableIpAddressCount'],
                sn['AvailabilityZone']
            ])

        print(table.get_string())



from prettytable import PrettyTable
from .utils import sizeof_fmt


class ContainerRegistry:
    def __init__(self, client):
        self.client = client

    def get_images(self, repo_name, account_id):
        if repo_name != '':
            response = self.client.list_images(
                registryId=account_id,
                repositoryName=repo_name,
                maxResults=123,
                filter={
                    'tagStatus': 'ANY'
                }
            )

            table = PrettyTable()
            table.field_names = ['Tag', 'Digest', 'Size']
            table.align['Tag'] = 'l'
            table.align['Digest'] = 'l'

            for img in response['imageIds']:
                data = self.get_image_metadata(repo_name, account_id, img['imageDigest'])

                table.add_row([
                    img['imageTag'],
                    img['imageDigest'],
                    sizeof_fmt(data['imageSizeInBytes'])
                ])

            print(table.get_string())

    def get_image_metadata(self, repo_name, account_id, image_digest):
        response = self.client.describe_images(
            registryId=account_id,
            repositoryName=repo_name,
            imageIds=[
                {
                    'imageDigest': image_digest,
                },
            ],
            filter={
                'tagStatus': 'ANY'
            }
        )

        return response['imageDetails'][0]

    def get_repositories(self):
        response = self.client.describe_repositories()

        table = PrettyTable()
        table.field_names = ['Name', 'Url']
        table.align['Name'] = 'r'
        table.align['Url'] = 'l'

        for repo in response['repositories']:
            table.add_row([
                repo['repositoryName'],
                repo['repositoryUri']
            ])

        print(table.get_string())



from .utils import Output, expand_filters, calculate_age
from prettytable import PrettyTable


class AwsResourceQuery:
    def __init__(self, client):
        self.client = client

    def list_ec2(self, name_filter=None):
        table = PrettyTable()
        table.field_names = self.get_table_header()
        table.align['PrivateIP'] = 'r'
        table.align['PublicIP'] = 'r'
        table.align['Name'] = 'r'
        table.align['Age'] = 'r'

        ec2_instances = self.list_ec2_instances(name_filter)
        if len(ec2_instances) > 0:
            table.add_rows(ec2_instances)
            print(table.get_string())

    @staticmethod
    def get_table_header():
        return ["ID", "Name", "Type", "PublicIP", "PrivateIP", "Age", "State"]

    @staticmethod
    def get_instance_name_from_tags(tags):
        for item in tags:
            if item['Key'] == 'Name':
                return item['Value']

        return '---'

    def list_ec2_instances(self, name_filter=None):
        query_filters = expand_filters(name_filter)
        instances = []
        try:
            resp = self.client.describe_instances()
        except Exception as e:
            Output.error(e)
            return []

        for reservation in resp['Reservations']:
            for instance in reservation['Instances']:
                instance_name = ''
                if 'Tags' in instance:
                    instance_name = self.get_instance_name_from_tags(instance['Tags'])

                instance_state = instance['State']['Name']

                if name_filter:
                    # search by instance name
                    if 'name' in query_filters and query_filters['name'] not in instance_name.lower():
                        continue
                    if 'state' in query_filters and query_filters['state'] not in instance_state.lower():
                        continue

                data = {
                    'Id': instance['InstanceId'],
                    'Name': instance_name,
                    'Type': instance['InstanceType'],
                    'PublicIp': instance['PublicIpAddress'] if 'PublicIpAddress' in instance else '',
                    'PrivateIp': instance['PrivateIpAddress'] if 'PrivateIpAddress' in instance else '',
                    'Age': calculate_age(instance['LaunchTime']),
                    'State': instance_state,
                }

                instances.append(data.values())

        return instances



from prettytable import PrettyTable
from .utils import Output
from .aws_config import AwsConfigManager


class Route53():
    @staticmethod
    def list_r53(client, zone_id='', name_filter=None):
        if zone_id != '':
            Output.header(f'List Records for ZoneID: {zone_id}')
            resp = client.list_resource_record_sets(
                HostedZoneId=zone_id
            )

            table = PrettyTable()
            table.field_names = ["Name", "Type", "Targets"]
            table.align['Name'] = 'r'
            table.align['Targets'] = 'l'

            for rec in resp['ResourceRecordSets']:
                record_name = rec['Name'].strip('.')
                if name_filter:
                    if name_filter not in record_name:
                        continue
                if 'AliasTarget' in rec:
                    table.add_row([
                        record_name,
                        rec['Type'],
                        '(alias) ' + rec['AliasTarget']['DNSName'].strip('.')[:128]
                    ])

                if 'ResourceRecords' in rec:
                    table.add_row([
                        record_name,
                        rec['Type'],
                        '\n'.join([d['Value'][:128] for d in rec['ResourceRecords']])
                    ])

            print(table.get_string())
            return None

        try:
            resp = client.list_hosted_zones(MaxItems="100")

            table = PrettyTable()
            table.field_names = ["Domain", "Id", "Records"]
            table.align['Domain'] = 'r'
            table.align['Records'] = 'r'

            if len(resp['HostedZones']) > 0:
                for zone in resp['HostedZones']:
                    table.add_row([
                        zone['Name'].strip('.'),
                        zone['Id'].replace('/hostedzone/', ''),
                        zone['ResourceRecordSetCount']
                    ])
                print(f'\nAccount id: {AwsConfigManager.account_id()}')
                print(table.get_string())
            else:
                Output.error(f'No hosted zones in account: {AwsConfigManager.account_id()}')
        except Exception as e:
            Output.error(e)



import os
from datetime import datetime, timezone


def sizeof_fmt(num, suffix="B"):
    for unit in ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"]:
        if abs(num) < 1024.0:
            return f"{num:3.1f}{unit}{suffix}"
        num /= 1024.0

    return f"{num:.1f}Yi{suffix}"


def calculate_age(date_time_string: datetime):
    current_time = datetime.now(timezone.utc)
    delta = int((current_time - date_time_string).total_seconds())

    return convert_seconds_to_age(delta)


def convert_seconds_to_age(delta):
    output = ''
    enough_info = 0

    # show weeks
    if delta >= (3600 * 24 * 7):
        new_delta = delta / (3600 * 24 * 7)
        delta = delta % (3600 * 24 * 7)
        output += f'{int(new_delta)}w'
        enough_info += 1
    # show days
    if delta >= (3600 * 24):
        new_delta = delta / (3600 * 24)
        delta = delta % (3600 * 24)
        output += f'{int(new_delta)}d'
        enough_info += 1
        if enough_info >= 2:
            return output
    # show hours
    if delta >= 3600:
        new_delta = delta / 3600
        delta = delta % 3600
        output += f'{int(new_delta)}h'
        enough_info += 1
        if enough_info >= 2:
            return output
    # show minutes
    if delta >= 60:
        new_delta = delta / 60
        delta = delta % 60
        output += f'{int(new_delta)}m'
        enough_info += 1
        if enough_info >= 2:
            return output

    if delta > 0:
        output += f'{int(delta)}s'

    return output


def expand_filters(name_filter):
    if name_filter == '' or name_filter is None:
        return {}

    elements = name_filter.split(',')
    filters = {}
    for element in elements:
        key, value = element.split('=')
        filters[str(key).lower()] = value

    return filters


class IniUtils:
    @staticmethod
    def check_directory_exists(file_path):
        os.makedirs(file_path, exist_ok=True)


class Output:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

    @staticmethod
    def header(text):
        print(f'{Output.HEADER}{text}{Output.ENDC}')

    @staticmethod
    def success(text):
        print(f'{Output.OKGREEN}{text}{Output.ENDC}')

    @staticmethod
    def error(text):
        print(f'{Output.FAIL}{text}{Output.ENDC}')
